/*
** Module   : Functions.java
** Abstract : Dispatcher methods for Java language, server-side database
**            functions which replace Progress built-in functions
**
** Copyright (c) 2004-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20060811   @28584 Created initial version. Backing methods for
**                           Progress built-in functions executed within
**                           the database server as custom, Java-language
**                           functions.
** 002 ECF 20060818   @28692 Added new functions.
** 003 GES 20060907   @29295 Added new variants of toString() that take
**                           format strings.
** 004 ECF 20060908   @29382 Reimplemented all toString() methods. Old
**                           implementations did not properly handle null
**                           values as unknown.
** 005 ECF 20060912   @29479 Added toInt() variant which accepts Integer.
**                           Changed Number variant to accept Double
**                           instead.
** 006 ECF 20061023   @30618 Fixed variants of toString() which do not
**                           take a format phrase. These were generating
**                           left-padded strings, but should generate
**                           unpadded strings.
** 007 ECF 20070309   @32368 Fixed round() method signatures. These were
**                           still using Number and int; need to use
**                           Double and Integer, respectively.
** 008 ECF 20070426   @33230 Added concat() method.
** 009 ECF 20070608   @34014 Replaced all Double with BigDecimal. Reflects
**                           new implementations of decimal and MathOps
**                           classes.
** 010 ECF 20070618   @34092 Added a variant of toDec() which accepts a
**                           BigDecimal value and a precision. Also
**                           reimplemented several methods which use
**                           BigDecimal to preserve precision.
** 011 ECF 20071210   @36494 Implemented error handling for CAN-FINDs.
**                           Built-in functions/operators within converted
**                           CAN-FIND statements must handle errors by
**                           returning null instead of throwing an
**                           exception. These are wrapped in the
**                           ErrorHandler.checkError() function, which
**                           checks pending error state and returns an
**                           appropriate result.
** 012 ECF 20080313   @37511 Expanded API to provide variants with
**                           exhaustive combinations of numeric parameter
**                           types. This is necessary to enable "manual"
**                           overloading user defined functions.
** 013 GES 20080910   @39765 Minor javadoc update.
** 014 GES 20081003   @40031 Matched signature changes.
** 015 CA  20090702   @43026 Added versions of toInt, toString and toDec which
**                           receive a java.sql.Date as a parameter.
** 016 ECF 20090721   @43314 Partially backed out #015 (@43026). Removed the
**                           Date variants of toInt() and toDec(). These were
**                           redundant with julianDayInt() and julianDayDec(),
**                           respectively.
** 017 ECF 20091002   @44178 Added reportPrecisionScale() helper function to
**                           report precision and scale of a numeric value.
** 018 GES 20110712          Javadoc fix.
** 019 AIL 20121206          Update from character to TextOps.
** 020 GES 20130224          Renamed julianDayDec/julianDayInt to toDec/toInt
**                           which should be equivalent.
** 021 ECF 20130314          Added trim function.
** 022 CS  20130315          Added support for: to-rowid, length, weekday, chr,
*                            date, caps, left-trim, maximum, minimum, logical, 
*                            num-entries and replace functions

*/

package com.goldencode.p2j.persist.pl;

import java.math.*;
import java.sql.Date;
import com.goldencode.p2j.util.*;

/**
 * A collection of static methods which provide the backing functionality for
 * Progress-like built-in functions, which are executed within the database
 * server as custom, Java-language functions.  This provides a much
 * faster alternative to client-side where clause processing (see {@link
 * com.goldencode.p2j.persist.WhereExpression WhereExpression} and its
 * related classes).  This mechanism (or client-side where clause processing)
 * must be used in cases where a function converted from Progress code must
 * retain semantics specific to the Progress environment.  Executing these
 * operations on the server is much preferred, since the avoidance of moving
 * records across the network only to test them for inclusion in a result set
 * is prohibitively expensive.
 * <p>
 * This class serves primarily as a consolidation point and dispatcher for
 * worker methods located in diverse classes.  All method calls made by the
 * database server are redirected to those worker methods.  However, since
 * the database server knows nothing of the P2J wrapper types used by those
 * methods to represent primitives, this class also performs a transformation
 * to (on input) and from (on output) the P2J wrapper types.  Thus, all
 * parameters and return types for methods in this class are specified in
 * terms of the standard J2SE wrapper types or primitive types.
 * <p>
 * The transformation handles conversion between <code>null</code> J2SE
 * wrapper types and the implementation of <code>unknown value</code> in the
 * P2J wrapper types.  This is possible because unknown value maps to SQL
 * NULL in the backing database.  However, certain data can be lost in the
 * transformation, and if these data are necessary, this mechanism should be
 * avoided.  Instead, client-side where clause processing should be used in
 * these cases.  Specifically, the following data are not represented by J2SE
 * wrappers or primitives and thus cannot be preserved during transformation:
 * <ul>
 *   <li>non-default precision of a <code>decimal</code> value;
 *   <li>non-default case sensitivity of a <code>character</code> value;
 *   <li>non-default timezone of a <code>date</code> value.
 * </ul>
 *
 * @see  Operators
 */
public final class Functions
{
   /**
    * Test whether one string begins with another case insensitively.
    *
    * @see  com.goldencode.p2j.util.TextOps#begins(Text, character)
    */
   public static Boolean begins(String a, String b)
   {
      try
      {
         return TextOps.begins(new character(a),
                               new character(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Test whether one string begins with another, optionally taking case
    * into account.
    *
    * @see  com.goldencode.p2j.util.TextOps#begins(String, String, boolean)
    */
   public static Boolean begins(String a, String b, boolean caseSensitive)
   {
      try
      {
         return TextOps.begins(a, b, caseSensitive).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Returns the character value of a numeric expression which must be
    * between 1 and 65534 (inclusive).
    *
    * @see  com.goldencode.p2j.util.character#chr(int)
    */
   public static String chr(Integer value)
   {
      try
      {
         return character.chr(new integer(value)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Test for a pattern match within a string case insensitively.
    *
    * @see  com.goldencode.p2j.util.TextOps#matches(Text, character)
    */
   public static Boolean matches(String a, String b)
   {
      try
      {
         return TextOps.matches(new character(a),
                                new character(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Test for a pattern match within a string, optionally taking case into
    * account.
    *
    * @see  com.goldencode.p2j.util.TextOps#matches(String, String, boolean)
    */
   public static Boolean matches(String a, String b, boolean caseSensitive)
   {
      try
      {
         return TextOps.matches(a, b, caseSensitive).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Retrieve the 1-based index at which one string is found within another,
    * or 0 if the string is not found.
    *
    * @see  com.goldencode.p2j.util.TextOps#indexOf(Text, Text)
    */
   public static Integer indexOf(String source, String target)
   {
      try
      {
         return TextOps.indexOf(new character(source),
                                new character(target)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Retrieve the 1-based index at which one string is found within another,
    * or 0 if the string is not found, starting at a particular index.
    *
    * @see  com.goldencode.p2j.util.TextOps#indexOf(Text, Text, NumberType)
    */
   public static Integer indexOf(String source, String target, Integer start)
   {
      try
      {
         return TextOps.indexOf(new character(source),
                                new character(target),
                                new integer(start)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Retrieve the 1-based index at which one string is found within another,
    * or 0 if the string is not found, starting at a particular index.
    *
    * @see  com.goldencode.p2j.util.TextOps#indexOf(Text, Text, NumberType)
    */
   public static Integer indexOf(String source,
                                 String target,
                                 BigDecimal start)
   {
      try
      {
         return TextOps.indexOf(new character(source),
                                new character(target),
                                new integer(start)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Retrieve the 1-based index at which one string is found within another,
    * or 0 if the string is not found, starting at a particular index,
    * optionally taking case into account.
    *
    * @see  com.goldencode.p2j.util.TextOps#indexOf(
    *          java.lang.String,
    *          java.lang.String,
    *          double,
    *          boolean)
    */
   public static Integer indexOf(String source,
                                 String target,
                                 Integer start,
                                 boolean caseSensitive)
   {
      try
      {
         return TextOps.indexOf(source,
                                target,
                                start,
                                caseSensitive).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Retrieve the 1-based index at which one string is found within another,
    * or 0 if the string is not found, starting at a particular index,
    * optionally taking case into account.
    *
    * @see  com.goldencode.p2j.util.TextOps#indexOf(
    *          java.lang.String,
    *          java.lang.String,
    *          double,
    *          boolean)
    */
   public static Integer indexOf(String source,
                                 String target,
                                 BigDecimal start,
                                 boolean caseSensitive)
   {
      try
      {
         return TextOps.indexOf(source,
                                target,
                                start.intValue(),
                                caseSensitive).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Transform the given value into an <code>integer</code>, then return it
    * as a J2SE <code>Integer</code>, or <code>null</code> if unknown.
    *
    * @see  com.goldencode.p2j.util.integer#integer(java.lang.Boolean)
    */
   public static Integer toInt(Boolean value)
   {
      try
      {
         return (new integer(value)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Transform the given value into an <code>integer</code>, then return it
    * as a J2SE <code>Integer</code>, or <code>null</code> if unknown.
    *
    * @see  com.goldencode.p2j.util.integer#integer(java.lang.Number)
    */
   public static Integer toInt(Integer value)
   {
      try
      {
         return (new integer(value)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Transform the given value into an <code>integer</code>, then return it
    * as a J2SE <code>Integer</code>, or <code>null</code> if unknown.
    *
    * @see  com.goldencode.p2j.util.integer#integer(java.lang.Number)
    */
   public static Integer toInt(BigDecimal value)
   {
      try
      {
         return (new integer(value)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Transform the given value into an <code>integer</code>, then return it
    * as a J2SE <code>Integer</code>, or <code>null</code> if unknown.
    *
    * @see  com.goldencode.p2j.util.integer#integer(java.lang.String)
    */
   public static Integer toInt(String value)
   {
      try
      {
         return (new integer(value)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Transform the given value into an <code>integer</code>, then return it
    * as a J2SE <code>Integer</code>, or <code>null</code> if unknown.
    *
    * @see  com.goldencode.p2j.util.integer#integer(java.util.Date)
    */
   public static Integer toInt(Date value)
   {
      try
      {
         return (new integer(value)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Transform the given value into a <code>decimal</code>, then return it
    * as a J2SE <code>BigDecimal</code>, or <code>null</code> if unknown.
    *
    * @see  com.goldencode.p2j.util.decimal#decimal(
    *          com.goldencode.p2j.util.logical)
    */
   public static BigDecimal toDec(Boolean value)
   {
      try
      {
         return (new decimal(new logical(value))).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Transform the given value into a <code>decimal</code>, then return it
    * as a J2SE <code>BigDecimal</code>, or <code>null</code> if unknown.
    *
    * @see  com.goldencode.p2j.util.decimal#decimal(java.lang.Number)
    */
   public static BigDecimal toDec(Integer value)
   {
      try
      {
         return (new decimal(new integer(value))).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Return the given <code>BigDecimal</code> value.  Do not alter the
    * precision.
    *
    * @param   value
    *          decimal numeric value.
    */
   public static BigDecimal toDec(BigDecimal value)
   {
      return value;
   }
   
   /**
    * Return the given <code>BigDecimal</code> value after applying the given
    * precision.  This function is provided for the special case of a literal
    * value, where the default precision derived from that literal is not
    * appropriate.
    *
    * @param   value
    *          Decimal numeric value.
    * @param   precision
    *          Precision to apply.
    */
   public static BigDecimal toDec(BigDecimal value, Integer precision)
   {
      try
      {
         return value.setScale(precision.intValue(),
                               BigDecimal.ROUND_HALF_UP);
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Transform the given value into a <code>decimal</code>, then return it
    * as a J2SE <code>BigDecimal</code>, or <code>null</code> if unknown.
    *
    * @see  com.goldencode.p2j.util.decimal#decimal(java.lang.String)
    */
   public static BigDecimal toDec(String value)
   {
      try
      {
         return (new decimal(value)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Transform the given value into a <code>decimal</code>, then return it
    * as a J2SE <code>BigDecimal</code>, or <code>null</code> if unknown.
    *
    * @see  com.goldencode.p2j.util.decimal#decimal(
    *          com.goldencode.p2j.util.date)
    */
   public static BigDecimal toDec(Date value)
   {
      try
      {
         return (new decimal(new date(value))).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Transform the given value into an <code>character</code>, then return
    * it as a J2SE <code>String</code>, or <code>null</code> if unknown.
    *
    * @see  com.goldencode.p2j.util.character#valueOf(boolean)
    */
   public static String toString(Boolean value)
   {
      try
      {
         return character.valueOf(new logical(value)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Transform the given value into an <code>character</code>, then return
    * it as a J2SE <code>String</code>, or <code>null</code> if unknown.
    *
    * @see  com.goldencode.p2j.util.character#valueOf(boolean,String)
    */
   public static String toString(Boolean value, String fmt)
   {
      try
      {
         return character.valueOf(new logical(value), fmt).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Transform the given value into an <code>character</code>, then return
    * it as a J2SE <code>String</code>, or <code>null</code> if unknown.
    *
    * @see  com.goldencode.p2j.util.character#valueOf(int)
    */
   public static String toString(Integer value)
   {
      try
      {
         return character.valueOf(new integer(value)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Transform the given value into an <code>character</code>, then return
    * it as a J2SE <code>String</code>, or <code>null</code> if unknown.
    *
    * @see  com.goldencode.p2j.util.character#valueOf(int,String)
    */
   public static String toString(Integer value, String fmt)
   {
      try
      {
         return character.valueOf(new integer(value), fmt).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Transform the given value into an <code>character</code>, then return
    * it as a J2SE <code>String</code>, or <code>null</code> if unknown.
    *
    * @see  com.goldencode.p2j.util.character#valueOf(BaseDataType)
    */
   public static String toString(Date value)
   {
      try
      {
         return character.valueOf(new date(value)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Transform the given value into an <code>character</code>, then return
    * it as a J2SE <code>String</code>, or <code>null</code> if unknown.
    *
    * @see  com.goldencode.p2j.util.character#valueOf(BaseDataType,String)
    */
   public static String toString(Date value, String fmt)
   {
      try
      {
         return character.valueOf(new date(value), fmt).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Transform the given value into an <code>character</code>, then return
    * it as a J2SE <code>String</code>, or <code>null</code> if unknown.
    *
    * @see  com.goldencode.p2j.util.character#valueOf(double)
    */
   public static String toString(BigDecimal value)
   {
      try
      {
         return character.valueOf(new decimal(value)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Transform the given value into an <code>character</code>, then return
    * it as a J2SE <code>String</code>, or <code>null</code> if unknown.
    *
    * @see  com.goldencode.p2j.util.character#valueOf(double,String)
    */
   public static String toString(BigDecimal value, String fmt)
   {
      try
      {
         return character.valueOf(new decimal(value), fmt).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Simply return the given string.  Normally, the value would be converted
    * to a <code>character</code>, then back to a J2SE <code>String</code>,
    * but this method short-circuits that unnecessary, round trip.
    *
    * @param   value
    *          String to be returned.
    *
    * @return  <code>value</code> parameter, unchanged.
    */
   public static String toString(String value)
   {
      try
      {
         return character.valueOf(new character(value)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Transform the given value into an <code>character</code>, then return
    * it as a J2SE <code>String</code>, or <code>null</code> if unknown.
    *
    * @see  com.goldencode.p2j.util.character#valueOf(String,String)
    */
   public static String toString(String value, String fmt)
   {
      try
      {
         return character.valueOf(new character(value), fmt).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Returns the Rowid representation of a String value
    *
    * @see  com.goldencode.p2j.util.rowid#parseRowidString
    */
   public static Long toRowid(String text)
   {
      // trim the text before returning it
      try
      {
         return rowid.parseRowidString(text);
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Concatenate two strings using P2J character's default behavior.  Return
    * the result as a J2SE <code>String</code>, or <code>null</code> if
    * unknown.
    *
    * @see  com.goldencode.p2j.util.TextOps#concat
    */
   public static String concat(String a, String b)
   {
      try
      {
         return TextOps.concat(new character(a),
                               new character(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Extract a substring from a string, starting at a specific index.
    *
    * @see  com.goldencode.p2j.util.TextOps#substring
    */
   public static String substringOf(String text, Integer pos)
   {
      try
      {
         return TextOps.substring(new character(text),
                                  new integer(pos)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Extract a substring of a specific length from a string, starting at a
    * specific index.
    *
    * @see  com.goldencode.p2j.util.TextOps#substring
    */
   public static String substringOf(String text, Integer pos, Integer len)
   {
      try
      {
         return TextOps.substring(new character(text),
                                  new integer(pos),
                                  new integer(len)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Extract a substring of a specific length and Progress type from a
    * string, starting at a specific index.
    *
    * @see  com.goldencode.p2j.util.TextOps#substring
    */
   public static String substringOf(String text,
                                    Integer pos,
                                    Integer len,
                                    String type)
   {
      try
      {
         return TextOps.substring(new character(text),
                                  new integer(pos),
                                  new integer(len),
                                  new character(type)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Returns the maximum from a list of values.
    * 
    * @param  vars 
    *         The variable list of arguments.
    * 
    * @return The maximum from the list of arguments
    */
   public static BigDecimal maximum(BigDecimal var1, BigDecimal var2)
   {
      try
      {
         return decimal.maximum(new decimal(var1), new decimal(var2)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);

         return null;
      }
   }
   
   /**
    * Returns the maximum from a list of values.
    * 
    * @param  vars 
    *         The variable list of arguments.
    * 
    * @return The maximum from the list of arguments
    */
   public static String maximum(String var1, String var2)
   {
      try
      {
         return TextOps.maximum(new character(var1), new character(var2)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Returns the maximum from a list of values.
    * 
    * @param  vars 
    *         The variable list of arguments.
    * 
    * @return The maximum from the list of arguments
    */
   public static Date maximum(Date var1, Date var2)
   {
      try
      {
         return new Date(date.maximum(new date(var1), new date(var2)).dateValue().getTime());
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Returns the minimum from a list of values.
    * 
    * @param  vars 
    *         The variable list of arguments.
    * 
    * @return The minimum from the list of arguments
    */
   public static BigDecimal minimum(BigDecimal var1, BigDecimal var2)
   {
      try
      {
         return decimal.minimum(new decimal(var1), new decimal(var2)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);

         return null;
      }
   }
   
   /**
    * Returns the minimum from a list of values.
    * 
    * @param  vars 
    *         The variable list of arguments.
    * 
    * @return The minimum from the list of arguments
    */
   public static String minimum(String var1, String var2)
   {
      try
      {
         return TextOps.minimum(new character(var1), new character(var2)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Returns the minimum from a list of values.
    * 
    * @param  vars 
    *         The variable list of arguments.
    * 
    * @return The minimum from the list of arguments
    */
   public static Date minimum(Date var1, Date var2)
   {
      try
      {
         return var1.compareTo(var2) < 0 ? var1 : var2;
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);

         return null;
      }
   }
   
   /**
    * Returns logical representation of a char variable from a format of type "true/false".
    * 
    * @param  var
    *         The variable we want to match against the logical format
    * @param  format
    *         The format of form "leftLogical/rightLogical" in which leftLogical will represent
    *         <code>true</code> and rightLogical represents <code>false</code>
    * @return The logical representation of a char variable from a format of type "true/false"
    */
   public static Boolean toLogical(String var, String format)
   {
      try
      {
         return new logical(var,format).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Replaces the given values into the destination.
    *
    * @see  @see  com.goldencode.p2j.util.TextOps#replaceAll
    */
   public static String replace(String source, String toReplace, String replacement)
   {
      try
      {
         return TextOps.replaceAll(source, toReplace, replacement).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Trim white space from the left side of text.
    *
    * @see  com.goldencode.p2j.util.TextOps#trim
    */
   public static String ltrimWS(String text)
   {
      try
      {
         return TextOps.leftTrim(text).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Trim white space from the left and right side of text.
    *
    * @see  com.goldencode.p2j.util.TextOps#trim
    */
   public static String trimWS(String text)
   {
      try
      {
         return TextOps.trim(text).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Returns the uppercase representation of the string
    *
    * @see  com.goldencode.p2j.util.TextOps#toUpperCase(String, boolean)
    */
   public static String caps(String text)
   {
      try
      {
         return TextOps.toUpperCase(text, true);
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Retrieve the number of entries in the variable.
    *
    * @see  com.goldencode.p2j.util.TextOps#numEntries(Text)
    */
   public static Integer numEntries(String value)
   {
      try
      {
         return TextOps.numEntries(value).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Retrieve the number of entries in the variable.
    *
    * @see  com.goldencode.p2j.util.TextOps#numEntries(Text, String)
    */
   public static Integer numEntries(String value, String separator)
   {
      try
      {
         return TextOps.numEntries(value, separator).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Get the length of a string.
    *
    * @see  com.goldencode.p2j.util.TextOps#length(Text)
    */
   public static Integer lengthOf(String text)
   {
      try
      {
         return TextOps.length(new character(text)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Get the length of a string.
    *
    * @see  com.goldencode.p2j.util.TextOps#length(Text)
    */
   public static Integer lengthOf(String text, String type)
   {
      try
      {
         return TextOps.length(new character(text), new character(type)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Extract a string from a comma-delimited list represented by another
    * string, given its index within the list.
    *
    * @see  com.goldencode.p2j.util.TextOps#entry
    */
   public static String entry(Integer i, String list)
   {
      try
      {
         return TextOps.entry(new integer(i),
                              new character(list)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Extract a string from a delimited list represented by another string,
    * given its index within the list and the delimiter.
    *
    * @see  com.goldencode.p2j.util.TextOps#entry
    */
   public static String entry(Integer i, String list, String delimiter)
   {
      try
      {
         return TextOps.entry(new integer(i),
                              new character(list),
                              new character(delimiter)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Look up the index of a string which represents an element in a
    * comma-delimited list represented by another string.
    *
    * @see  com.goldencode.p2j.util.TextOps#lookup(character, Text)
    */
   public static Integer lookup(String entry, String list)
   {
      try
      {
         return TextOps.lookup(new character(entry),
                               new character(list)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Look up the index of a string which represents an element in a
    * delimited list represented by another string, using a specific
    * delimiter.
    *
    * @see  com.goldencode.p2j.util.TextOps#lookup(
    *          com.goldencode.p2j.util.character,
    *          com.goldencode.p2j.util.Text,
    *          com.goldencode.p2j.util.character)
    */
   public static Integer lookup(String entry, String list, String delimiter)
   {
      try
      {
         return TextOps.lookup(new character(entry),
                               new character(list),
                               new character(delimiter)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Look up the index of a string which represents an element in a
    * delimited list represented by another string, using a specific
    * delimiter, and optionally taking case into account.
    *
    * @see  com.goldencode.p2j.util.TextOps#lookup(
    *          java.lang.String,
    *          java.lang.String,
    *          char,
    *          boolean)
    */
   public static Integer lookup(String entry,
                                String list,
                                String delimiter,
                                boolean caseSensitive)
   {
      try
      {
         return TextOps.lookup(entry,
                               list,
                               delimiter.charAt(0),
                               caseSensitive).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Determine whether a comma-separated list contains a specific item.
    *
    * @see  com.goldencode.p2j.util.TextOps#matchesList(Text, character)
    */
   public static Boolean matchesList(String list, String item)
   {
      try
      {
         return TextOps.matchesList(new character(list),
                                    new character(item)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Round a floating point number to 10 decimal places.
    *
    * @see  com.goldencode.p2j.util.MathOps#round(
    *          com.goldencode.p2j.util.decimal)
    */
   public static BigDecimal roundDec(BigDecimal value)
   {
      try
      {
         return MathOps.round(value);
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Round a floating point number to a particular precision.
    *
    * @see  com.goldencode.p2j.util.MathOps#round(
    *          com.goldencode.p2j.util.decimal,
    *          int)
    */
   public static BigDecimal roundDec(BigDecimal value, Integer precision)
   {
      try
      {
         return MathOps.round(value, precision.intValue());
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Retrieve the date from a string variable.
    * 
    * @param  text
    *         The text value from which we return the new date.
    *         
    * @return The calculated date value.
    */
   public static Date toDate(String text)
   {
      try
      {
         return new Date(new date(text).dateValue().getTime());
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Retrieve the date from a string variable.
    * 
    * @param  value
    *         The integer representing the number of days from the Progress 4GL base data time.
    *         
    * @return The calculated date value.
    */
   public static Date toDate(Integer value)
   {
      try
      {
         return new Date(new date(value).dateValue().getTime());
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Retrieve the date from a string variable.
    * 
    * @param  month
    *         The month of the date.
    *         day
    *         The day of the date.
    *         year
    *         The year of the date.
    *         
    * @return The calculated date value
    */
   public static Date toDate(Integer month, Integer day, Integer year)
   {
      try
      {
         return new Date(new date(new integer(month), 
                                  new integer(day), 
                                  new integer(year)).dateValue().getTime());
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Retrieve the week day number from a date.
    *
    * @see  com.goldencode.p2j.util.date#weekday(com.goldencode.p2j.util.date)
    */
   public static Integer toWeekday(Date value)
   {
      try
      {
         return date.weekday(new date(value)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Retrieve the year number from a date.
    *
    * @see  com.goldencode.p2j.util.date#year(com.goldencode.p2j.util.date)
    */
   public static Integer getYear(Date value)
   {
      try
      {
         return date.year(new date(value)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Retrieve the month number (1-based) from a date.
    *
    * @see  com.goldencode.p2j.util.date#month(com.goldencode.p2j.util.date)
    */
   public static Integer getMonth(Date value)
   {
      try
      {
         return date.month(new date(value)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Retrieve the day of the month (1-based) from a date.
    *
    * @see  com.goldencode.p2j.util.date#year(com.goldencode.p2j.util.date)
    */
   public static Integer getDay(Date value)
   {
      try
      {
         return date.day(new date(value)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Helper function to report the precision and scale of a numeric value.
    * 
    * @param   num
    *          Numeric value.
    * 
    * @return  String representing the precision <i>p</i> and the scale
    *          <i>s</i> in the form <code>(<i>p</i>,<i>s</i>)</code>.
    */
   public static String reportPrecisionScale(BigDecimal num)
   {
      StringBuilder buf = new StringBuilder();
      buf.append("(");
      buf.append(num.precision());
      buf.append(",");
      buf.append(num.scale());
      buf.append(")");
      
      return buf.toString();
   }
}
